USE master
GO

IF OBJECT_ID('sp_ABBackupDb') IS NULL
    EXEC('CREATE PROCEDURE sp_ABBackupDb AS BEGIN RETURN END')
GO


ALTER PROCEDURE sp_ABBackupDb (@Databases varchar(1000) = '',    -- Can be a comma separated list of databases
                                                                 --   or {SystemDbs}, {UserDbs}, or {AllDbs}
                               @FileSpec varchar(1024) = '',     -- Can contain wildcards, chooses the latest matching
                                                                 -- The & character is replaced with current db name
                                                                 -- The @ character is replaced with YYYYMMDDHHMM.
                                                                 -- If empty, uses default backup location from registry,
                                                                 --   appends directory names if @CreateDir = 1 (default),
                                                                 --   then uses format "&_db_@", extension SQB or BAK
                                                                 -- Multiple filenames can be comma-separated
                               @Options varchar(400) = '',       -- Additional WITH options to append after the backup string
                                                                  --  Lightspeed proc options can be included, e.g. INIT, @Logging=1
                                                                  --  becomes @With='...,INIT',@Logging=1
                               @Type varchar(10) = '',           -- SQL = Native SQL Server
                                                                 -- SQB = Sql Backup
                                                                 -- Lightspeed = SQL Litespeed
                                                                 -- If empty, determines appropriate based on presence of XPs.
                               @Verify int = 0,                  -- If 1, verifies the backup file
                               @DeleteAfter varchar(10) = '',    -- If @ in filespec, delete files older than this interval
                                                                 -- Should be a number followed by 'hours', 'days', or 'weeks'
                               @CreateDir int = 1,               -- Create directory if necessary
                               @DryRun int = 0)                  -- If 1, don't do backup

/* Name:    sp_ABBackupDb
 *
 * Purpose: Backs up and optionally verifies a list of databases using either native SQL Server commands, 
 *          SQL Litespeed, or RedGate's SQL Backup.  The filespec can contain & (replaced by database name) 
 *          and wildcards.  The most recent file matching the spec is used as the filename.  Multiple filenames 
 *          can be comma-separated.
 *
 * Examples: EXEC sp_ABBackupDb 'Northwind, pubs', 'E:\MSSQL\BACKUP\&\&_db_@.BAK'
 *
 * Returns: 0 on success, otherwise maximum error number encountered.  Note that native SQL error will likely blow up the
 *          batch and not return to the caller.
 *
 * Modification History:
 *  09/29/2005  VRI  Created.
 *  10/24/2005  VRI  Added @CreateDir parameter.
 *
 */

AS

SET NOCOUNT ON
SET ANSI_NULLS ON

DECLARE @Sql varchar(2000),
        @AllOpts varchar(4000),
        @Q char(1),
        @InstanceName varchar(30),
        @Database varchar(80),
        @StartTime datetime,
        @ElapsedMin int,
        @ElapsedSec int,
        @File varchar(1000),
        @Id int,
        @Error int, 
        @VerError int,
        @MaxError int,
        @DateStr char(15),
        @regkey varchar(1024),
        @BackupDir varchar(1024),
        @Ext char(3),
        @Cmd varchar(1024),
        @Path varchar(200),
        @FileClause varchar(2000),
        @FileId int,
        @Pos int,
        @DeleteHours int,
        @LightspeedParms varchar(1000)


CREATE TABLE #FileExist (FileFound int,
                         DirFound int, 
                         ParentDirFound int)

CREATE TABLE #Dir(LineNum int IDENTITY PRIMARY KEY, 
                  Line varchar(200) NULL)

DECLARE @FileNames TABLE (FileName varchar(1000), 
                          Id int)

SET @Q = ''''
SET @MaxError = 0

-- Do a cursory check for SQL Injection inside our parameter strings
SET @AllOpts = ' ' + @Options + ' ' + @Filespec + ' ' + @Databases
IF @AllOpts LIKE '% SELECT %'
OR @AllOpts LIKE '% UPDATE %'
OR @AllOpts LIKE '% DELETE %'
OR @AllOpts LIKE '% INSERT %'
OR @AllOpts LIKE '% TRUNCATE %'
OR @AllOpts LIKE '% EXEC %'
OR @AllOpts LIKE '% EXECUTE %'
OR @AllOpts LIKE '%--%'
OR @AllOpts LIKE '%;%'
BEGIN
    RAISERROR ('Parameters rejected - potential SQL Injection attempt.', 16, 1)
    RETURN(-1)
END


IF @@SERVERNAME LIKE '%\%'
BEGIN
    SELECT @InstanceName = substring(@@SERVERNAME, charindex('\', @@SERVERNAME) + 1, 30)
    SELECT @regkey       = 'SOFTWARE\Microsoft\Microsoft SQL Server\' + @InstanceName + '\MSSQLServer'
END
ELSE
    SELECT @InstanceName = '', 
           @regkey = 'SOFTWARE\Microsoft\MSSQLServer\MSSQLServer'

EXEC master.dbo.xp_regread 'HKEY_LOCAL_MACHINE',  @regkey, 'BackupDirectory',  @param = @BackupDir OUTPUT 
    
-- Expand {UserDbs},{SystemDbs},{AllDbs}, or an empty database name param to the current db
IF @Databases LIKE '%{%}%'
BEGIN
    IF @Databases LIKE '%{AllDbs}%'
        SET @Databases = '{UserDbs},{SystemDbs}'

    IF @Databases LIKE '%{UserDbs}%'
    BEGIN
        DECLARE @DbStr varchar(1000)
        SELECT  @DbStr = ISNULL(@DbStr + ',', '') + rtrim(name)
          FROM  master.dbo.sysdatabases
          WHERE (@Databases LIKE '%{UserDbs}%'
            AND name NOT IN ('master', 'msdb', 'model', 'tempdb'))
          ORDER BY name
        SET @Databases = REPLACE(@Databases, '{UserDbs}', @DbStr)
    END  
    IF @Databases LIKE '%{SystemDbs}%'
        SET @Databases = REPLACE(@Databases, '{SystemDbs}', 'master,model,msdb')
END
ELSE IF @Databases = ''
    SET @Databases = DB_NAME()


IF ISNULL(@Options, '') = '' 
    SET @Options = 'INIT'

-- Determine restore type if not specified based on presence of extended stored procs
IF ISNULL(@Type,'') = ''
BEGIN
    IF EXISTS (SELECT  * 
                 FROM  master.dbo.sysobjects 
                 WHERE Type = 'X'
                   AND Name = 'xp_restore_database')
        SET  @Type = 'Lightspeed'
    ELSE IF EXISTS (SELECT  * 
                      FROM  master.dbo.sysobjects 
                      WHERE Type = 'X'
                        AND Name = 'sqlbackup')
        SET  @Type = 'SQB'
    ELSE
        SET @Type = 'SQL'
END

IF @Type = 'SQB'
    SET @Ext = 'SQB'
ELSE IF @Type IN ('SQL', 'Lightspeed')
    SET @Ext = 'BAK'
ELSE
BEGIN
    RAISERROR('Type %s is not one of SQL, Lightspeed, or SQB', 16, 1, @Type)
    RETURN -1
END

-- For Lightspeed only, separate parms in the Option string that start with @
IF @Type = 'Lightspeed' AND @Options LIKE '%@%=%'
BEGIN
    SELECT  SLSOption = ltrim(rtrim(Item)),
            ItemId
      INTO  #SLSOptions
      FROM  ::fn_Split(@Options, ',')

    SELECT  @LightspeedParms = ISNULL(@LightspeedParms + ', ','') + SLSOption
      FROM  #SLSOptions
      WHERE SLSOption LIKE '@%'
      ORDER BY ItemId

    SET @Options = NULL

    SELECT  @Options = ISNULL(@Options + ', ','') + SLSOption
      FROM  #SLSOptions
      WHERE SLSOption NOT LIKE '@%'
      ORDER BY ItemId
END                

-- Process the @DeleteAfter option
IF ISNULL(@DeleteAfter, '') <> '' 
BEGIN
    IF isnumeric(ltrim(left(@DeleteAfter, charindex(' ', @DeleteAfter)))) = 1
    BEGIN
        SET @DeleteHours = ltrim(left(@DeleteAfter, charindex(' ', @DeleteAfter)))
        IF @DeleteAfter LIKE '%hour%' 
            SET @DeleteHours = @DeleteHours * 1
        ELSE IF @DeleteAfter LIKE '%day%'
            SET @DeleteHours = @DeleteHours * 24
        ELSE IF @DeleteAfter LIKE '%week%'
            SET @DeleteHours = @DeleteHours * 24 * 7
        ELSE 
        BEGIN
            RAISERROR ('Ignoring invalid @DeleteAfter parameter %s', 0, 1, @DeleteAfter)
            SET @DeleteHours = 0
        END
    END
    ELSE
        RAISERROR ('Ignoring invalid @DeleteAfter parameter %s', 0, 1, @DeleteAfter)
END


    
-- Split the database string on a comma
SELECT  Db = ltrim(Item),
        Id = ItemId
  INTO  #Dbs
  FROM  ::fn_Split(@Databases, ',')
  JOIN  master.dbo.sysdatabases
    ON  Item = name

-- Figure out the @FileSpec if left blank.  Use default backup directory and, if @CreateDir is true include
-- the directory name, otherwise no directories.  If @FileSpec doesn't start with a drive or UNC name, prepend 
-- the default backup directory.
IF ISNULL(@FileSpec,'') = ''
    SET @FileSpec = CASE WHEN @CreateDir = 1 THEN '&\' ELSE '' END + '&_db_@.' + @Ext

-- Filespec can contain multiple files, so split them up...
SELECT  FileSpec = ltrim(Item),
        Id = ItemId
  INTO  #FileSpecs
  FROM  ::fn_Split(@FileSpec, ',')

-- Fill in the backup directory if relative path and the extension if not supplied
UPDATE #FileSpecs
  SET  FileSpec = @BackupDir + CASE WHEN LEFT(FileSpec,1) = '\'
                                    THEN ''
                                    ELSE '\'
                               END + FileSpec + CASE WHEN RIGHT(FileSpec,4) NOT LIKE '%.%'
                                                      AND RIGHT(FileSpec,1) NOT LIKE '%*%'
                                                     THEN '.' + @Ext
                                                     ELSE ''
                                                END
  WHERE FileSpec NOT LIKE '\\%' AND FileSpec NOT LIKE '_:%'

           
SELECT  @Id = 0

-- Do the backup for all databases passed in
WHILE EXISTS (SELECT  * 
                FROM  #Dbs 
                WHERE Id > @Id)
BEGIN

    SELECT  @Database = Db,
            @Id = Id
      FROM  #Dbs
      WHERE Id = @Id + 1


    SET @FileId = 0
    DELETE @FileNames

    -- Handle each file in the filespec
    WHILE EXISTS (SELECT  *
                    FROM  #FileSpecs
                    WHERE Id > @FileId)
    BEGIN

        SELECT  @File = FileSpec,
                @FileId = Id
          FROM  #FileSpecs
          WHERE Id = @FileId + 1

        -- If & appears in filespec, replace with current database name
        SELECT @File = REPLACE(@File, '&', @Database)
        SELECT @DateStr = REPLACE(REPLACE(REPLACE(CONVERT(char(16), GetDate(), 120), '-', ''), ':',''), ' ','')
        SELECT @File = REPLACE(@File, '@', @DateStr)
        SELECT @Path = substring(@File, 1, len(@file) - charindex('\', reverse(rtrim(@File))) + 1)

        -- If a wildcard pattern, choose the most recent filename matching the pattern
        IF @File LIKE '%*%' OR @File LIKE '%?%'
        BEGIN
            SET @Cmd = 'dir /b /o-d ' + @File

            DELETE #Dir

            INSERT #Dir (Line)
              EXEC master.dbo.xp_cmdshell @Cmd

            SELECT  TOP 1 @File = @Path + Line
              FROM  #Dir
              WHERE Line IS NOT NULL
                AND Line <> 'The system cannot find the file specified.'
                AND Line <> 'File Not Found'
              ORDER BY LineNum

            IF @@Rowcount = 0
            BEGIN
                RAISERROR ('Filespec %s did not match any files', 16, 1, @File)
                RETURN -1
            END
        END
        ELSE IF @CreateDir = 1
        BEGIN
            SET @Cmd = 'if not exist ' + @Path + ' mkdir ' + @Path
            EXEC @Error = master.dbo.xp_cmdshell @Cmd
            IF @Error <> 0
            BEGIN
                RAISERROR('Error %d creating directory path %s.', 16, 1, @Error, @Path)
                RETURN -1
            END
        END
                
        INSERT @FileNames (FileName, Id)
          SELECT @File, @FileId
    END

    -- Construct the backup statement
    SELECT  @FileClause = NULL
    SELECT  @FileClause = CASE WHEN @Type = 'Lightspeed' 
                               THEN ISNULL(@FileClause + ',', '') + '@FileName='
                               ELSE ISNULL(@FileClause + ',', '') + 'DISK=' 
                          END + @Q + FileName + @Q
      FROM  @FileNames
      ORDER BY Id

    IF @Type = 'SQL'
        SELECT @Sql = 'BACKUP DATABASE ' + @Database + ' TO ' + @FileClause + ' WITH STATS = 10, ' + @Options
    ELSE IF @Type = 'Lightspeed'
        SELECT @Sql = 'master.dbo.xp_backup_database @database = ' + @Q + @Database + @Q + ', ' + @FileClause + 
                    + ', @With=''Stats=10, ' + REPLACE(@Options, @Q, @Q+@Q) + '''' + ISNULL(@LightspeedParms,'')
    ELSE IF @Type = 'SQB'
        SELECT @Sql = 'master.dbo.sqlbackup ''-E -I ' + @InstanceName + ' -SQL "BACKUP DATABASE ' + @Database 
                    + ' TO ' + REPLACE(@FileClause, @Q, @Q+@Q)
                    + ' WITH ' + REPLACE(@Options, @Q, @Q+@Q) + '"'''

    PRINT @Sql
    SET @StartTime = getdate()

    IF @DryRun = 0
    BEGIN
        -- Run the backup
        TRUNCATE TABLE #Dir
        SET @Error = 0
        SET @StartTime = getdate()

        -- We can't do a native backup/restore inside a transaction, so
        -- simply look for an error number coming back.  Most errors will
        -- simply blow up the batch in a way we can't catch, though.
        IF @Type = 'SQL'
        BEGIN
            SET @Sql = 'USE master ' + @Sql
            EXEC (@Sql)
            SET @Error = @@Error
        END
        -- For Lightspeed we select the error code into the table
        ELSE IF @Type = 'Lightspeed'
        BEGIN 
            SET @Sql = 'USE master DECLARE @ret int EXEC @ret = ' + @Sql + ' SELECT @ret' 
            INSERT #Dir
              EXEC (@Sql)

            SELECT  TOP 1 @Error = convert(int, Line)
              FROM  #Dir
              ORDER BY LineNum
        END
        -- SQL Backup does table output, so look for Msg xxxx as the error
        ELSE
        BEGIN
            SET @Sql = 'USE master'
                     + ' DECLARE @ret int'
                     + ' EXEC @ret = ' + @Sql  
                     + ' IF @Ret = 0 RAISERROR(''SQL Backup Restore Failed!'',16,1)'
            INSERT #Dir
              EXEC (@Sql)

            SELECT  Line 
              FROM  #Dir 
              ORDER BY LineNum

            SELECT  TOP 1 @Error = convert(int, SUBSTRING(Line, 5, charindex(',', Line) - 5))
              FROM  #Dir
              WHERE Line LIKE 'Msg [0-9]%'
              ORDER BY LineNum
        END
        IF @Error > @MaxError
            SET @MaxError = @Error

        SELECT  @ElapsedMin = DateDiff(second, @StartTime, getdate()) / 60,
                @ElapsedSec = DateDiff(second, @StartTime, getdate()) % 60
        RAISERROR('Database %s backed up in %02d:%02d', 0, 1, @Database, @ElapsedMin, @ElapsedSec)
    END

    IF @Verify = 1
        BEGIN
        -- Construct the verify statement
        IF @Type = 'SQL'
            SELECT @Sql = 'RESTORE VERIFYONLY FROM ' + @FileClause
        ELSE IF @Type = 'Lightspeed'
            SELECT @Sql = 'master.dbo.xp_restore_verifyonly ' + @FileClause 
                        + ' IF @Ret <> 0 RAISERROR(''SQL Lightspeed Verify Failed (return code %d)!'',16,1,@ret)'
        ELSE IF @Type = 'SQB'
            SELECT @Sql = 'master.dbo.sqlbackup ''-E -I ' + @InstanceName + ' -SQL "RESTORE VERIFYONLY FROM ' + REPLACE(@FileClause, @Q, @Q+@Q) + '"'''
                        + ' IF @Ret = 0 RAISERROR(''SQL Backup Verify Failed!'',16,1)'

        PRINT @Sql
        SET @StartTime = getdate()

        IF @DryRun = 0
        BEGIN
            -- Run the verify
            TRUNCATE TABLE #Dir
            SET @VerError = 0
            SET @StartTime = getdate()

            -- We can't do a native backup/restore inside a transaction, so
            -- simply look for an error number coming back.  Most errors will
            -- simply blow up the batch in a way we can't catch, though.
            IF @Type = 'SQL'
            BEGIN
                SET @Sql = 'USE master ' + @Sql
                EXEC (@Sql)
                SET @VerError = @@Error
            END
            -- For Lightspeed we select the error code into the table
            ELSE IF @Type = 'Lightspeed'
            BEGIN 
                SET @Sql = 'USE master DECLARE @ret int EXEC @ret = ' + @Sql + ' SELECT @ret' 
                INSERT #Dir
                  EXEC (@Sql)

                SELECT  TOP 1 @VerError = convert(int, Line)
                  FROM  #Dir
                  ORDER BY LineNum
            END
            -- SQL Backup does table output, so look for Msg xxxx as the error
            ELSE
            BEGIN
                SET @Sql = 'USE master'
                         + ' DECLARE @ret int'
                         + ' EXEC @ret = ' + @Sql  
                         + ' IF @Ret = 0 RAISERROR(''SQL Backup Restore Failed!'',16,1)'
                INSERT #Dir
                  EXEC (@Sql)

                SELECT  Line 
                  FROM  #Dir 
                  ORDER BY LineNum

                SELECT  TOP 1 @VerError = convert(int, SUBSTRING(Line, 5, charindex(',', Line) - 5))
                  FROM  #Dir
                  WHERE Line LIKE 'Msg [0-9]%'
                  ORDER BY LineNum
            END
            IF @VerError > @MaxError
                SET @MaxError = @VerError
        END
    END

    -- Remove old files if @DeleteAfter parm included
    IF @DeleteHours > 0 
    BEGIN
        SET @FileId = 0

        -- Handle each file in the filespec
        WHILE EXISTS (SELECT  *
                        FROM  #FileSpecs
                        WHERE Id > @FileId)
        BEGIN

            SELECT  @File = FileSpec,
                    @FileId = Id
              FROM  #FileSpecs
              WHERE Id = @FileId + 1

            -- Only delete if the filespec has a date
            IF @File NOT LIKE '%@%'
                CONTINUE

            -- If & appears in filespec, replace with current database name
            SELECT @File = REPLACE(@File, '&', @Database)
            SELECT @Pos = charindex('@', @File)
            SELECT @File = REPLACE(@File, '@', '*')

            SET @Path = substring(@File, 1, len(@file) - charindex('\', reverse(rtrim(@File))) + 1)
            SET @Cmd = 'dir /b /o-d ' + @File

            DELETE #Dir

            INSERT #Dir (Line)
              EXEC master.dbo.xp_cmdshell @Cmd

            -- remove non-dates
            DELETE  #Dir
              WHERE isdate(stuff(stuff(substring(@Path + Line, @Pos, 12), 9, 0, ' '), 12, 0, ':')) <> 1
                OR  Line IS NULL

            -- remove if less than threshhold
            SELECT  @Path + Line AS FileName,
                    LineNum,
                    DATEDIFF(hour,  
                             stuff(stuff(substring(@Path + Line, @Pos, 12), 9, 0, ' '), 12, 0, ':'), 
                             stuff(stuff(@DateStr, 9, 0, ' '), 12, 0, ':')) AS HourInterval
              INTO  #ToDelete      
              FROM  #Dir

            DELETE  #ToDelete
              WHERE HourInterval < @DeleteHours

            -- Run a DOS delete command on whichever filenames are left
            WHILE EXISTS (SELECT  * 
                            FROM  #ToDelete)
            BEGIN
                SELECT  TOP 1 @File = FileName
                  FROM  #ToDelete
                  ORDER BY LineNum

                DELETE  #ToDelete
                  WHERE @File = FileName

                IF @Error > 0
                BEGIN
                    RAISERROR ('NOT removing old file %s since error %d was returned from backup.', 0, 1, @File, @Error)
                    CONTINUE
                END

                IF @VerError > 0
                BEGIN
                    RAISERROR ('NOT removing old file %s since error %d was returned from verify.', 0, 1, @File, @VerError)
                    CONTINUE
                END

                RAISERROR ('Removing old file %s', 0, 1, @File)

                SET @Cmd = 'del ' + @File
                IF @DryRun = 0
                    EXEC master.dbo.xp_cmdshell @Cmd, no_output
            END
            DROP TABLE #ToDelete
        END
    END
END

RETURN(@MaxError)
GO

GRANT EXEC ON sp_ABBackupDb TO PUBLIC
GO