USE master
GO

IF OBJECT_ID('sp_ABRestoreDb') IS NULL
    EXEC('CREATE PROCEDURE sp_ABRestoreDb AS BEGIN RETURN END')
GO


ALTER PROCEDURE sp_ABRestoreDb (@Databases varchar(1000),         -- Can be a comma separated list of databases
                                                                  -- Can be {UserDbs},{SystemDbs},{AllDbs} (master excluded)
                                @FileSpec varchar(400) = '',      -- Can contain wildcards, chooses the latest matching
                                                                  -- The & character is replaced with current db name
                                                                  -- If blank, chooses the backup directory in the registry
                                                                  --   plus '\&\&*' if exists, '\&*' otherwise
                                                                  --   relative paths prepended with default backup directory
                                                                  -- Extension of .BAK or .SQB supplied added if not specified
                                                                  -- Multiple filenames can be comma-separated
                                @Options varchar(400) = '',       -- Additional restore options
                                                                  --  Lightspeed proc options can be included, e.g. RECOVERY, @Logging=1
                                                                  --  becomes @With='MOVE ...,RECOVERY',@Logging=1
                                @Type varchar(10) = '',           -- SQL = Native SQL Server
                                                                  -- SQB = Sql Backup 
                                                                  -- Litespeed = SQL Litespeed
                                                                  -- If blank, determines based on extended stored procs in master
                                @DryRun int = 0)                  -- If 1, don't run the restore

/* Name:    sp_ABRestoreDb
 *
 * Purpose: Restore a list of databases using either native SQL Server commands, SQL Litespeed, or RedGate's SQL Backup.  
 *          First kick users off database, then restore, then fixup user login links.  The filespec can contain & (replaced
 *          by database name) and wildcards.  The most recent file matching the spec is used as source of input.
 *
 *          If the database to restore exists and is online, and the logical names of the files to restore match, the existing
 *          database filenames are used rather than the filenames in the backup file.  
 *
 *          If the database doesn't exist or is offline (e.g. loading, suspect), the default data and log paths are used 
 *          (or else the SQL Server data directory + '/DATA').  If the logical names of the file to restore and the database
 *          don't match but have the same number and type, best guess matches to the logical filenames are made based on file
 *          type (data and log) and size of file.
 *          
 *
 * Examples: -- restore Northwind and pubs from the latest files like E:\MSSQL\BACKUP\Northwind\Northwind_db_*.BAK
 *           --  and E:\MSSQL\BACKUP\pubs\pubs_db_*.BAK.  Uses SQL Backup or Lightspeed if installed.
 *           EXEC sp_ABRestoreDb 'Northwind, pubs', 'E:\MSSQL\BACKUP\&\&_db_*.BAK'  
 *
 *           -- restore latest from E:\MSSQL\BACKUP\Northwind\Northwind* or else E:\MSSQL\BACKUP\Northwind* 
 *           -- Assumes SQL Backup directory is set to E:\MSSQL\BACKUP.  Uses SQL Backup or Lightspeed if installed.
 *           EXEC sp_ABRestoreDb 'Northwind'
 *
 *           -- restore latest from E:\MSSQL\BACKUP\Northwind\Northwind_db_* with NORECOVERY option.  Use native SQL.
 *           EXEC sp_ABRestoreDb 'Northwind', 'E:\MSSQL\BACKUP\&\&_db_*', 'NORECOVERY', 'SQL'
 *
 *           -- create database newpubs on the default data and log locations from pubs dump.
 *           EXEC sp_ABRestoreDb 'newpubs', 'E:\MSSQL\BACKUP\pubs.BAK'
 *
 * Returns:  0 on success, otherwise maximum error number encountered.  Note that native SQL errors will likely blow up
 *           the batch and not return to the caller.
 *
 * Modification History:
 *  08/22/2005  VRI  Created.
 *  08/31/2005  VRI  Added wildcard filespecs to the @File parameter.
 *  09/06/2005  VRI  Added instance name checking for Type = SQB.
 *  09/14/2005  VRI  Allow list of databases and & replacement in filename
 *  09/29/2005  VRI  Fill in a blank FileSpec and add options
 *  10/05/2005  VRI  Handle nonexistant/offline database and non-matching logical filenames.
 *  10/07/2005  VRI  Error checking and multiple filenames.
 *
 */

AS

SET NOCOUNT ON
SET ANSI_NULLS ON

DECLARE @Sql varchar(2000),
        @AllOpts varchar(4000),
        @Logical varchar(100),
        @Physical varchar(1000),
        @Move varchar(3000),
        @Q char(1),
        @InstanceName varchar(30),
        @Database varchar(80),
        @StartTime datetime,
        @ElapsedMin int,
        @ElapsedSec int,
        @File varchar(1000),
        @Id int,
        @Error int, 
        @MaxError int,
        @Ext char(3),
        @regkey varchar(1024),
        @setupkey varchar(1024),
        @BackupDir varchar(1024),
        @Cmd varchar(1024),
        @Path varchar(512),
        @FirstLine int,
        @DefaultData varchar(512),
        @DefaultLog varchar(512),
        @Dcount int,
        @Lcount int,
        @FileClause varchar(2000),
        @FileId int,
        @LightspeedParms varchar(1000),
        @Pos int,
        @Parm varchar(1000)

CREATE TABLE #FileList (LogicalName nvarchar(128),                 -- Logical file name. 
                        fileid smallint,                           -- Numeric file identifier. 
                        PhysicalName nvarchar(260),                -- Operating-system file name (physical file name). 
                        FileGroup nvarchar(128),                   -- Group in which the file belongs. 
                        Size nvarchar(18),                         -- File size. 
                        MaxSize nvarchar(18),                      -- Maximum size to which the file can grow. 
                        Growth nvarchar(18),                       -- Growth increment of the file. 
                        Usage varchar(9)                           -- 'Data only' or 'Log only'
                       )

CREATE TABLE #FileListOutput (LogicalName nvarchar(128),           -- Logical name of the file 
                              PhysicalName nvarchar(260),          -- Physical or operating-system name of the file 
                              Type char(1),                        -- Data file (D) or a log file (L) 
                              FileGroupName nvarchar(128) NULL,    -- Name of the filegroup that contains the file 
                              Size numeric(20,0) NULL,             -- Current size in bytes 
                              MaxSize numeric(20,0) NULL,          -- Maximum allowed size in bytes 
                              RowNum int identity(1,1))          

CREATE TABLE #Dir(LineNum int IDENTITY PRIMARY KEY, 
                  Line varchar(200) NULL)

DECLARE @FileList1 TABLE (PhysicalName nvarchar(260),
                          Type char(1),
                          MyOrder int IDENTITY(1,1))

DECLARE @FileListOutput1 TABLE (LogicalName nvarchar(260),
                                Type char(1),
                                MyOrder int IDENTITY(1,1))

DECLARE @FileNames TABLE (FileName varchar(1000), 
                          Id int)


SET @Q = ''''
SET @MaxError = 0

-- Do a cursory check for SQL Injection inside our parameter strings
SET @AllOpts = ' ' + @Options + ' ' + @Filespec + ' ' + @Databases
IF @AllOpts LIKE '% SELECT %'
OR @AllOpts LIKE '% UPDATE %'
OR @AllOpts LIKE '% DELETE %'
OR @AllOpts LIKE '% INSERT %'
OR @AllOpts LIKE '% TRUNCATE %'
OR @AllOpts LIKE '% EXEC %'
OR @AllOpts LIKE '% EXECUTE %'
OR @AllOpts LIKE '%--%'
OR @AllOpts LIKE '%;%'
BEGIN
    RAISERROR ('Parameters rejected - potential SQL Injection attempt.', 16, 1)
    RETURN(-1)
END

IF @@SERVERNAME LIKE '%\%'
BEGIN
    SELECT @InstanceName = substring(@@SERVERNAME, charindex('\', @@SERVERNAME) + 1, 30)
    SELECT @regkey       = 'SOFTWARE\Microsoft\Microsoft SQL Server\' + @InstanceName + '\MSSQLServer',
           @setupkey     = 'SOFTWARE\Microsoft\Microsoft SQL Server\' + @InstanceName + '\Setup'
END
ELSE
    SELECT @InstanceName = '', 
           @regkey   = 'SOFTWARE\Microsoft\MSSQLServer\MSSQLServer',
           @setupkey = 'SOFTWARE\Microsoft\MSSQLServer\Setup'


EXEC master.dbo.xp_regread 'HKEY_LOCAL_MACHINE',  @regkey, 'BackupDirectory',  @param = @BackupDir OUTPUT 
    
IF ISNULL(@Options, '') = '' 
    SET @Options = 'RECOVERY'
IF SUBSTRING(@Options,1,1) <> ','
    SET @Options = ',' + @Options

-- Expand {UserDbs},{SystemDbs},{AllDbs}, or an empty database name param to the current db
IF @Databases LIKE '%{%}%'
BEGIN
    IF @Databases LIKE '%{AllDbs}%'
        SET @Databases = '{UserDbs},{SystemDbs}'

    IF @Databases LIKE '%{UserDbs}%'
    BEGIN
        DECLARE @DbStr varchar(1000)
        SELECT  @DbStr = ISNULL(@DbStr + ',', '') + rtrim(name)
          FROM  master.dbo.sysdatabases
          WHERE (@Databases LIKE '%{UserDbs}%'
            AND name NOT IN ('master', 'msdb', 'model','tempdb'))
          ORDER BY name
        SET @Databases = REPLACE(@Databases, '{UserDbs}', @DbStr)
    END  
    IF @Databases LIKE '%{SystemDbs}%'
        SET @Databases = REPLACE(@Databases, '{SystemDbs}', 'model,msdb')  -- don't include master
END

-- Determine restore type if not specified based on presence of extended procs
IF ISNULL(@Type,'') = ''
BEGIN
    IF EXISTS (SELECT  * 
                 FROM  master.dbo.sysobjects 
                 WHERE Type = 'X'
                   AND Name = 'xp_restore_database')
        SET  @Type = 'Lightspeed'
    ELSE IF EXISTS (SELECT  * 
                      FROM  master.dbo.sysobjects 
                      WHERE Type = 'X'
                        AND Name = 'sqlbackup')
        SET  @Type = 'SQB'
    ELSE
        SET @Type = 'SQL'
END

IF @Type = 'SQB'
    SET @Ext = 'SQB'
ELSE IF @Type IN ('SQL','Lightspeed')
    SET @Ext = 'BAK'
ELSE
BEGIN
    RAISERROR('Type %s is not one of SQL, Lightspeed, or SQB', 16, 1, @Type)
    RETURN -1
END

-- For Lightspeed only, separate parms in the Option string that start with @
IF @Type = 'Lightspeed' AND @Options LIKE '%@%=%'
BEGIN
    SELECT  SLSOption = ltrim(rtrim(Item)),
            ItemId
      INTO  #SLSOptions
      FROM  ::fn_Split(@Options, ',')

    SELECT  @LightspeedParms = ISNULL(@LightspeedParms + ', ','') + SLSOption
      FROM  #SLSOptions
      WHERE SLSOption LIKE '@%'
      ORDER BY ItemId

    SET @Options = NULL

    SELECT  @Options = ISNULL(@Options + ', ','') + SLSOption
      FROM  #SLSOptions
      WHERE SLSOption NOT LIKE '@%'
      ORDER BY ItemId
END                
    
-- Split the database string on a comma
SELECT  Db = ltrim(Item),
        Id = ItemId
  INTO  #Dbs
  FROM  ::fn_Split(@Databases, ',')

-- Figure out the @FileSpec if left blank.  Use default backup directory and, if all databases to be restored
-- have directories with the name of the db, use the directories, otherwise no directories.
-- If @FileSpec doesn't start with a drive or UNC name, prepend the default backup directory
IF ISNULL(@FileSpec,'') = ''
BEGIN
    SELECT @Cmd = 'dir /b /ad ' + @BackupDir

    INSERT #Dir (Line)
        EXEC master.dbo.xp_cmdshell @Cmd

    IF EXISTS (SELECT  * 
                 FROM  #Dbs 
                 WHERE NOT EXISTS (SELECT  * 
                                     FROM #Dir 
                                     WHERE #Dbs.Db = #Dir.Line))
        SET @FileSpec = @BackupDir + '\&_db_*.' + @Ext
    ELSE 
        SET @FileSpec = @BackupDir + '\&' + '\&_db_*.' + @Ext
END

-- Filespec can contain multiple files, so split them up...
SELECT  FileSpec = ltrim(Item),
        Id = ItemId
  INTO  #FileSpecs
  FROM  ::fn_Split(@FileSpec, ',')

-- Fill in the backup directory if relative path and the extension if not supplied
UPDATE #FileSpecs
  SET  FileSpec = @BackupDir + CASE WHEN LEFT(FileSpec,1) = '\'
                                    THEN ''
                                    ELSE '\'
                               END + FileSpec + CASE WHEN RIGHT(FileSpec,4) NOT LIKE '%.%'
                                                      AND RIGHT(FileSpec,1) NOT LIKE '%*%'
                                                     THEN '.' + @Ext
                                                     ELSE ''
                                                END
  WHERE FileSpec NOT LIKE '\\%' AND FileSpec NOT LIKE '_:%'

SELECT  @Id = 0

-- Do the restore for all databases passed in
WHILE EXISTS (SELECT  * 
                FROM  #Dbs 
                WHERE Id > @Id)
BEGIN

    SELECT  @Database = Db,
            @Id = Id
      FROM  #Dbs
      WHERE Id = @Id + 1

    -- If the database to restore is our current database, we have to skip it 
    IF @Database = DB_NAME()
    BEGIN
        RAISERROR ('Cannot restore current database %s', 0, 1, @Database)
        GOTO NextDb
    END

    SET @FileId = 0
    DELETE @FileNames

    -- Handle each file in the filespec
    WHILE EXISTS (SELECT  *
                    FROM  #FileSpecs
                    WHERE Id > @FileId)
    BEGIN

        SELECT  @File = FileSpec,
                @FileId = Id
          FROM  #FileSpecs
          WHERE Id = @FileId + 1

        -- If & appears in filespec, replace with current database name
        SELECT @File = REPLACE(@File, '&', @Database)

        -- If a wildcard pattern, choose the most recent filename matching the pattern
        IF @File LIKE '%*%' OR @File LIKE '%?%'
        BEGIN
            SET @Path = substring(@File, 1, len(@file) - charindex('\', reverse(rtrim(@File))) + 1)
            SET @Cmd = 'dir /b /o-d ' + @File

            DELETE #Dir

            INSERT #Dir (Line)
              EXEC master.dbo.xp_cmdshell @Cmd

            SELECT  TOP 1 @File = @Path + Line
              FROM  #Dir
              WHERE Line IS NOT NULL
                AND Line <> 'The system cannot find the file specified.'
                AND Line <> 'File Not Found'
              ORDER BY LineNum

            IF @@Rowcount = 0
            BEGIN
                RAISERROR ('Filespec %s did not match any files', 16, 1, @File)
                RETURN -1
            END
        END

        -- Put our completed filenames into a table
        INSERT @FileNames (FileName, Id)
          SELECT @File, @FileId
    END

    -- Read the logical and physical devices from the file to load
    DELETE #FileListOutput

    IF @Type = 'SQB'
    BEGIN
        SELECT @Cmd = 'exec master.dbo.sqlbackup ''-I ' + @InstanceName + ' -E -SQL "RESTORE FILELISTONLY FROM DISK=''''' + @File + '''''"'''

        TRUNCATE TABLE #Dir

        INSERT #Dir (Line)
            EXEC (@Cmd)

        -- Ignore the first few lines of input
        SELECT  @FirstLine = MIN(LineNum)
          FROM  #Dir
          WHERE Line LIKE 'Logical name%'

        -- Copy the output to the table, stopping at the C-style 0 byte, in sets of 4 lines
        INSERT #FileListOutput (LogicalName,
                                PhysicalName,
                                Type)
            SELECT MAX(CASE WHEN Line LIKE 'Logical name%'
                            THEN substring(Line, charindex(':', Line) + 2, charindex(char(0), Line) - charindex(':', Line) - 2)
                            ELSE ''
                       END),
                   MAX(CASE WHEN Line LIKE 'Physical name%'
                            THEN substring(Line, charindex(':', Line) + 2, charindex(char(0), Line) - charindex(':', Line) - 2)
                            ELSE ''
                       END), 
                   MAX(CASE WHEN Line LIKE 'Type%'
                            THEN substring(Line, charindex(':', Line) + 2, 1)
                            ELSE ''
                       END) 
              FROM #Dir
              WHERE LineNum >= @FirstLine
              GROUP BY (LineNum - @FirstLine) / 4 
    END
    ELSE 
    BEGIN 
        IF @Type = 'SQL' 
            SELECT @Sql = 'RESTORE FILELISTONLY FROM DISK=''' + @File + ''''
        ELSE IF @Type = 'Lightspeed'
            SELECT @Sql = 'EXEC master.dbo.xp_restore_filelistonly @filename=''' + @File + ''''

        INSERT #FileListOutput
          EXEC (@Sql)
    END


    -- Does the database to restore exist and is it online?  Then read the file structure...
    IF convert(varchar(255), DATABASEPROPERTYEX(@Database, 'Status')) LIKE '%ONLINE%'
    BEGIN
        -- Get the file list on the existing database we will restore to
        SELECT @Sql = 'USE ' + @Database + ' EXEC sp_helpfile',
               @Move = NULL

        DELETE #FileList

        INSERT #FileList
          EXEC (@Sql)

        -- If the logical names all match, use the existing filenames from the database to restore
        IF NOT EXISTS (SELECT  * 
                         FROM  #FileList
                         FULL  OUTER JOIN #FileListOutput
                           ON  #FileList.LogicalName = #FileListOutput.LogicalName
                         WHERE (#FileList.PhysicalName IS NULL
                            OR  #FileListOutput.PhysicalName IS NULL))
        BEGIN
            SELECT  @Move = ISNULL(@Move + ', ' , '') + 'MOVE ' + @Q + rtrim(LogicalName) + @Q 
                            + ' TO ' + @Q + rtrim(PhysicalName) + @Q
              FROM  #FileList
              ORDER BY fileid,
                       LogicalName
        END
        -- If the logical names don't match, but the same number of each type exists, match by type and size
        ELSE IF  (SELECT  COUNT(*) FROM  #FileList WHERE LEFT(Usage,1) = 'D') = (SELECT  COUNT(*) FROM  #FileListOutput WHERE Type = 'D')
             AND (SELECT  COUNT(*) FROM  #FileList WHERE LEFT(Usage,1) = 'L') = (SELECT  COUNT(*) FROM  #FileListOutput WHERE Type = 'L')
        BEGIN

            INSERT  @FileList1 (PhysicalName,
                                Type)
              SELECT  rtrim(FL.PhysicalName),
                      left(FL.Usage,1)
                FROM  #FileList FL
                ORDER BY FL.Usage,
                         Right(rtrim(FL.PhysicalName), 3),
                         FL.Size desc

            INSERT  @FileListOutput1 (LogicalName,
                                      Type)
              SELECT  rtrim(FLO.LogicalName),
                      FLO.Type
                FROM  #FileListOutput FLO
                ORDER BY FLO.Type,
                         Right(rtrim(FLO.PhysicalName), 3),
                         FLO.Size desc

            SELECT  @Move = ISNULL(@Move + ', ' , '') + 'MOVE ' + @Q + LogicalName + @Q + ' TO ' + @Q + PhysicalName + @Q
              FROM  @FileList1 FL1      
              JOIN  @FileListOutput1 FLO1
                ON  FL1.MyOrder = FLO1.MyOrder
              ORDER BY FL1.MyOrder
        END
        ELSE
        BEGIN
            RAISERROR('Couldn''t restore database %s from file %s since the number/type of logical devices do not match.', 16, 1, @Database, @File)
            SET @MaxError = 50010
            GOTO NextDb
        END
    END
    ELSE   -- Database doesn't exist or is not online
    BEGIN
        -- Get the default data/log locations
        EXEC master.dbo.xp_regread 'HKEY_LOCAL_MACHINE',  @regkey, 'DefaultData',  @param = @DefaultData OUTPUT 
        EXEC master.dbo.xp_regread 'HKEY_LOCAL_MACHINE',  @regkey, 'DefaultLog',   @param = @DefaultLog OUTPUT 

        -- If not there, use the SQL Server setup location for SQL Data
        IF @DefaultData IS NULL
        BEGIN
            EXEC master.dbo.xp_regread 'HKEY_LOCAL_MACHINE',  @setupkey, 'SQLDataRoot',  @param = @DefaultData OUTPUT 
            IF @DefaultData NOT LIKE '%\'
                SET @DefaultData = @DefaultData + '\'
            SET @DefaultData = @DefaultData + 'Data\'
        END
        IF @DefaultLog IS NULL
            SET @DefaultLog = @DefaultData

        IF @DefaultData NOT LIKE '%\'
            SET @DefaultData = @DefaultData + '\'
        IF @DefaultLog NOT LIKE '%\'
            SET @DefaultLog = @DefaultLog + '\'
 
        SELECT  @Move = NULL

        SELECT  @Move = ISNULL(@Move + ', ' , '') + 'MOVE ' + @Q + rtrim(LogicalName) + @Q + ' TO ' 
                        + @Q + CASE WHEN Type = 'D' 
                                    THEN @DefaultData + @Database + '_data' 
                                         + CASE WHEN PhysicalName LIKE '%.MDF' THEN '.MDF' ELSE convert(varchar(2), isnull(@Dcount,0)+1) + '.NDF' END
                                    WHEN Type = 'L'
                                    THEN @DefaultLog  + @Database + '_log' 
                                         + CASE WHEN @Lcount >= 1 THEN convert(varchar(2), @Lcount+1) ELSE '' END + '.LDF'
                               END + @Q,
                @DCount = isnull(@Dcount,0) + CASE WHEN Type = 'D' THEN 1 ELSE 0 END,
                @LCount = isnull(@Lcount,0) + CASE WHEN Type = 'L' THEN 1 ELSE 0 END
          FROM  #FileListOutput

    END
        
    -- Construct the restore statement
    SELECT  @FileClause = NULL
    SELECT  @FileClause = CASE WHEN @Type = 'Lightspeed' 
                               THEN ISNULL(@FileClause + ',', '') + '@FileName='
                               ELSE ISNULL(@FileClause + ',', '') + 'DISK=' 
                          END + @Q + FileName + @Q
      FROM  @FileNames
      ORDER BY Id

    IF @Type = 'SQL'
        SELECT @Sql = 'RESTORE DATABASE ' + @Database + ' FROM ' + @FileClause + ' WITH STATS = 10, ' + @Move + @Options
    ELSE IF @Type = 'Lightspeed'
        SELECT @Sql = 'master.dbo.xp_restore_database @database = ' + @Q + @Database + @Q + ', ' + @FileClause 
                    + ', @With=''Stats=5, ' + REPLACE(@Move, @Q, @Q+@Q) + REPLACE(@Options, @Q, @Q+@Q) + '''' 
                    + ISNULL(@LightspeedParms, '')
    ELSE IF @Type = 'SQB'
        SELECT @Sql = 'master.dbo.sqlbackup ''-E -I ' + @InstanceName + ' -SQL "RESTORE DATABASE ' + @Database 
                    + ' FROM ' + REPLACE(@FileClause, @Q, @Q+@Q)
                    + ' WITH ' + REPLACE(@Move, @Q, @Q+@Q) + REPLACE(@Options, @Q, @Q+@Q) + '"'''

    PRINT @Sql

    IF @DryRun = 1
        GOTO NextDb

    -- Remove existing users in the database
    EXEC sp_abKillDbUsers @Database

    -- Run the restore
    TRUNCATE TABLE #Dir
    SET @Error = 0
    SET @StartTime = getdate()

    -- We can't do a native backup/restore inside a transaction, so
    -- simply look for an error number coming back.  Most errors will
    -- simply blow up the batch in a way we can't catch, though.
    IF @Type = 'SQL'
    BEGIN
        SET @Sql = 'USE master ' + @Sql
        EXEC (@Sql)
        SET @Error = @@Error
    END
    -- For Lightspeed we select the error code into the table
    ELSE IF @Type = 'Lightspeed'
    BEGIN 
        SET @Sql = 'USE master DECLARE @ret int EXEC @ret = ' + @Sql + ' SELECT @ret' 
        INSERT #Dir
          EXEC (@Sql)

        SELECT  TOP 1 @Error = convert(int, Line)
          FROM  #Dir
          ORDER BY LineNum
    END
    -- SQL Backup does table output, so look for Msg xxxx as the error
    ELSE
    BEGIN
        SET @Sql = 'USE master'
                 + ' DECLARE @ret int'
                 + ' EXEC @ret = ' + @Sql  
                 + ' IF @Ret = 0 RAISERROR(''SQL Backup Restore Failed!'',16,1)'
        INSERT #Dir
          EXEC (@Sql)

        SELECT  Line 
          FROM  #Dir 
          ORDER BY LineNum

        SELECT  TOP 1 @Error = convert(int, SUBSTRING(Line, 5, charindex(',', Line) - 5))
          FROM  #Dir
          WHERE Line LIKE 'Msg [0-9]%'
          ORDER BY LineNum
    END
    IF @Error > @MaxError
        SET @MaxError = @Error

    IF @Error > 0
        GOTO NextDb

    SELECT  @ElapsedMin = DateDiff(second, @StartTime, getdate()) / 60,
            @ElapsedSec = DateDiff(second, @StartTime, getdate()) % 60
    RAISERROR('Database %s restored in %02d:%02d', 0, 1, @Database, @ElapsedMin, @ElapsedSec)

    -- Now fixup the user login links
    SET @Sql = 'USE ' + @Database + ' EXEC sp_ABFixUserLogins'
    EXEC (@Sql)
    SET @Error = @@Error
    IF @Error > @MaxError
        SET @MaxError = @Error
NextDb:
END

RETURN(@MaxError)
GO

GRANT EXEC ON sp_ABRestoreDb TO PUBLIC
GO